
#include "GISproto.h"

#include <tk.h>
#include <stdlib.h>
#include <stdio.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "Event.h"

#define PROCMAX 3
#define CAMP_LEN 29 // 7*sizeof( float ) + 1 = 29

struct AppInfo
{
  struct GISproc_info procTab[PROCMAX];
  char camera[CAMP_LEN];
  float flyHeight;
  char* url;
};

static struct AppInfo GISstatus;


#define MST 1
#define SLV 2

static char app_mode;
static char* suffix[] = { ".sat", ".map", ".wth" };

static char buffer[URL_MAXLEN];
static char cbuf;
static int GISnumber;
static float height;

// CIV stuff
static int civ_sock;
static int AID;
static int port;
static int CIVheader[2];
// CIV stuff

char* trans_addr;
int init;

void
main( int argc, char*argv[] )
{
  if( argc > 2 )
    {
      AID = atoi( argv[1] );
      port = atoi( argv[2] );
#ifdef CIV_MODE      
      if( argc > 3 )
	{
	  fprintf( stderr, "less params - this is CIV_MODE\n" );
	  exit(1);
	}
#else
      if( argc > 3 )
	trans_addr = argv[3];
      else
	{
	  fprintf( stderr, "more params - this is *not* CIV_MODE\n" );
	  exit(1);
	}
#endif	  
      
      Tk_Main( 1, argv, Tcl_AppInit );

      exit(0);
    } 
  else
    {
      fprintf( stderr, "more params\n" );
      exit(1);
    }
}



void
readFromMasterBrowser( ClientData, int )
{
  float* cPrms = ( float* )( GISstatus.camera+1 );
  readAll( GISstatus.procTab[0].sockfd, GISstatus.camera, CAMP_LEN );
#ifdef DEBUG
  fprintf( stderr, "received CAMP:\t\t(%f, %f, %f)\n\t\t\t(%f, %f, %f, %f)\n",
	   cPrms[0], cPrms[1], cPrms[2], cPrms[3], cPrms[4], cPrms[5], cPrms[6] );
#endif
  if( app_mode == MST )
    {
      CIVheader[0] = DATA;
      CIVheader[1] = CAMP_LEN;
#ifdef DEBUG
      fprintf( stderr, "Cam - CIVheader: %d, %d\n", CIVheader[0], CIVheader[1] );
#endif
      write( civ_sock, CIVheader, sizeof( CIVheader ) );
      write( civ_sock, GISstatus.camera, CAMP_LEN );
    }
  for( int GISnumber=1; GISnumber<PROCMAX; GISnumber++ )
    if( GISstatus.procTab[GISnumber].stat != GIS_OFF )
      if( GISstatus.procTab[GISnumber].orient == FLY ) 
	{
	  height = cPrms[1];
	  cPrms[1] = GISstatus.flyHeight;
	  write( GISstatus.procTab[GISnumber].sockfd, GISstatus.camera, CAMP_LEN );
	  cPrms[1] = height;
	}
      else
	write( GISstatus.procTab[GISnumber].sockfd, GISstatus.camera, CAMP_LEN );
}

void
spawnGIS( int GISnumber, Tcl_Interp *interp )
{
  pid_t chldPid;
  int fd[2];
  socketpair( AF_UNIX, SOCK_STREAM, 0, fd ); 
  switch( chldPid = fork() ) 
    { 
    case 0 :
      close( fd[1] );
#ifdef DEBUG
      fprintf( stderr, "new proc no. %d with sockfd %d\n", getpid(), fd[0] );
#endif

      char strsock[8];
      sprintf( strsock, "%d", fd[0] );

      GISstatus.url = Tcl_GetVar( interp, "newurl", 0 );
      if( *GISstatus.url )	  
	{
	  strncpy( buffer, GISstatus.url, URL_MAXLEN-5 );
	  strcat( buffer,  GISstatus.procTab[GISnumber].cont == TERRAIN ? suffix[0]:
		  GISstatus.procTab[GISnumber].cont == MAP ? suffix[1] : suffix[2] );
	  execl( GISapp, "tweety", (GISnumber || app_mode == SLV)?"s":"m", 
		 GISstatus.procTab[GISnumber].orient == WALK ? "w" : "f",
		 strsock, buffer, NULL ); 
	} 
      else 
	execl( GISapp, "tweety", (GISnumber || app_mode == SLV)?"s":"m",
	       GISstatus.procTab[GISnumber].orient == WALK ? "w" : "f",
	       strsock, NULL ); 
      break;
    case -1: 
      perror( "spawnProc" ); 
      exit(1);
      break;
    default: 
      GISstatus.procTab[GISnumber].stat = GIS_ON;
      GISstatus.procTab[GISnumber].pid = chldPid;
      GISstatus.procTab[GISnumber].sockfd = fd[1];
	  
      if( GISnumber == 0 )
	Tk_CreateFileHandler( fd[1], TK_READABLE, readFromMasterBrowser, NULL );
      close( fd[0] );
    }
}

void
killGIS( int GISnumber )
{
  if( GISnumber == 0 )
    Tk_DeleteFileHandler( GISstatus.procTab[0].sockfd );
  else
    Tk_DeleteFileHandler( GISstatus.procTab[GISnumber].sockfd );
  kill( GISstatus.procTab[GISnumber].pid, SIGTERM );
  close( GISstatus.procTab[GISnumber].sockfd );
  GISstatus.procTab[GISnumber].sockfd = 0;
  GISstatus.procTab[GISnumber].pid = 0;
  GISstatus.procTab[GISnumber].stat = GIS_OFF;
}

const char*
compURL( const char* newURL, const int cont )
{
  size_t len;
  
  strncpy( &buffer[2], newURL, URL_MAXLEN-7 ); // 7 = 2(=&buffer[2] - &buffer[0] ) + 5(=".sat")
  buffer[0] = NURL;
  len = strlen( &buffer[2] );
  strcpy( &buffer[2] + len, suffix[cont] );
  buffer[1] = len + 5; //strlen + suffix(4) +'\0'
#ifdef DEBUG
  fprintf( stderr, "Content: %d, NURL: %s, len %d\n", cont, &buffer[2], buffer[1] );
#endif	    
  return buffer;
}

void
readFromContrApp( ClientData, int )
{
  readAll( civ_sock, CIVheader, sizeof( CIVheader ) );
#ifdef DEBUG
  fprintf( stderr, "CIVheader: %d, %d\n", CIVheader[0], CIVheader[1] );
#endif
  read( civ_sock, GISstatus.camera, 1 );
  switch( *GISstatus.camera )
    {
    case NOBJ:
      //      fprintf( stderr, "NOBJ received\n" );
      readAll( civ_sock, &buffer[2], CIVheader[1]-1 );
      buffer[0] = NOBJ;
      buffer[1] = CIVheader[1] + 1; // strlen + '\0'
      for( GISnumber=0; GISnumber<PROCMAX; GISnumber++ )
	if( GISstatus.procTab[GISnumber].stat != GIS_OFF )
	  {
#ifdef DEBUG
	    fprintf( stderr, "NOBJ: %s, len %d\n", &buffer[2], buffer[1] );
#endif	    
	    write( GISstatus.procTab[GISnumber].sockfd, buffer, buffer[1]+2 );
 	  }
      break;
      
    case SREQ:
      //      fprintf( stderr, "SYNC ask received\n" );
      CIVheader[0] = DATA;
      CIVheader[1] = 10;
      buffer[0] = SANS;
      for( GISnumber=0; GISnumber<PROCMAX; GISnumber++ )
	memcpy( &buffer[1] + GISnumber * 3, &GISstatus.procTab[GISnumber].stat, 3 );
      
      write( civ_sock, CIVheader, sizeof( CIVheader ) );
      write( civ_sock, buffer, CIVheader[1] );
      //      fprintf( stderr, "GISstatus sent\n" );

      if( GISstatus.url /*&& *GISstatus.url*/ )
	{
	  CIVheader[1] = strlen( GISstatus.url )+2; // NURL + strlen + '0'
 	  cbuf = NURL; 
#ifdef DEBUG 
	  fprintf( stderr, "CIVheader: %d, %d\n", CIVheader[0], CIVheader[1] ); 
#endif  
	  write( civ_sock, CIVheader, sizeof( CIVheader ) );  
	  write( civ_sock, &cbuf, 1 );  
	  write( civ_sock, GISstatus.url, CIVheader[1]-1 );  
	  //	  fprintf( stderr, "NURL sent\n" );  
	} 

      CIVheader[1] = 2;
      buffer[0] = CHGT;
      buffer[1] = ( char )GISstatus.flyHeight;
#ifdef DEBUG
      fprintf( stderr, "CIVheader: %d, %d\n", CIVheader[0], CIVheader[1] );
      fprintf( stderr, "Data: %d, %d\n", buffer[0], buffer[1] );
#endif
      write( civ_sock, CIVheader, sizeof( CIVheader ) );
      write( civ_sock, buffer, 2 );
      //      fprintf( stderr, "CHGT sent\n" );  

      CIVheader[1] = CAMP_LEN;
      GISstatus.camera[0] = CAMP; 
#ifdef DEBUG 
      fprintf( stderr, "Cam - CIVheader: %d, %d\n", CIVheader[0], CIVheader[1] ); 
#endif 
      write( civ_sock, CIVheader, sizeof( CIVheader ) ); 
      write( civ_sock, GISstatus.camera, CAMP_LEN ); 
      //      fprintf( stderr, "CAMP sent\n" );  
      break;
      
    default:
      ;
      //      fprintf( stderr, "MST: Control protocol violation: %d\t", *GISstatus.camera );
      //      fprintf( stderr, "pruning of next %d bytes\n", read( civ_sock, buffer, CIVheader[1]-1 ) );
    }
}

void
readFromMasterApp( ClientData interp, int )
{
  float* cPrms = ( float* )( GISstatus.camera+1 );
  readAll( civ_sock, CIVheader, sizeof( CIVheader ) );
#ifdef DEBUG
  fprintf( stderr, "CIVheader: %d, %d\n", CIVheader[0], CIVheader[1] );
#endif
  read( civ_sock, GISstatus.camera, 1 );
  switch( *GISstatus.camera )
    {
    case CAMP:
      readAll( civ_sock, GISstatus.camera+1 , CAMP_LEN-1 );
      for( GISnumber=0; GISnumber<PROCMAX; GISnumber++ )
	if( GISstatus.procTab[GISnumber].stat != GIS_OFF )
	  {
	    if( GISstatus.procTab[GISnumber].orient == FLY ) 
	      {
		height = cPrms[1];
		cPrms[1] = GISstatus.flyHeight;
		write( GISstatus.procTab[GISnumber].sockfd, GISstatus.camera, CAMP_LEN );
		cPrms[1] = height;
	      }
	    else
	      write( GISstatus.procTab[GISnumber].sockfd, GISstatus.camera, CAMP_LEN );
#ifdef DEBUG
	    fprintf( stderr, "---\nreceived CAMP:\t\t(%f, %f, %f)\n\t\t\t(%f, %f, %f, %f)\n",
		     cPrms[0], cPrms[1], cPrms[2], cPrms[3], cPrms[4], cPrms[5], cPrms[6] );
#endif
	  }
      break;

    case NURL:
      readAll( civ_sock, &buffer[2], CIVheader[1]-1 );
      // Move it down, next time !!!!!!
      Tcl_SetVar( ( Tcl_Interp* )interp, "newurl", &buffer[2], 0 );
      // end of move sect.
      if( ! buffer[2] )
	{
	  //	  fprintf( stderr, "empty newURL received\n" );
	  return;
	}
#ifdef DEBUG
      fprintf( stderr, "received NURL: %s\n", &buffer[2] );
#endif
      buffer[0] = NURL;
      cbuf = strlen( &buffer[2] );
      for( GISnumber=0; GISnumber<PROCMAX; GISnumber++ )
	if( GISstatus.procTab[GISnumber].stat != GIS_OFF )
	  {
	    strcpy( &buffer[2] + cbuf, GISstatus.procTab[GISnumber].cont == TERRAIN ? suffix[0]:
		    GISstatus.procTab[GISnumber].cont == MAP ? suffix[1] : suffix[2] );

	    buffer[1] = cbuf + 5; //strlen + suffix(4) +'\0'
#ifdef DEBUG
	    fprintf( stderr, "sending NURL: %s, len %d\n", &buffer[2], buffer[1] );
#endif	    
	    write( GISstatus.procTab[GISnumber].sockfd, buffer, buffer[1]+2 );
 	  }
      break;

    case CTER:
      readAll( civ_sock, buffer, CIVheader[1]-1 );
      GISnumber = buffer[0];
      if( GISstatus.procTab[GISnumber].stat != GIS_OFF )
	{
	  GISstatus.url = Tcl_GetVar( ( Tcl_Interp* )interp, "newurl", 0 );
	  if( *GISstatus.url )
	    {
	      compURL( GISstatus.url, TERRAIN );
	      write( GISstatus.procTab[GISnumber].sockfd, buffer, buffer[1]+2 );

	      GISstatus.camera[0] = CAMP;
	      if( GISstatus.procTab[GISnumber].orient == FLY ) 
		{
		  height = cPrms[1];
		  cPrms[1] = GISstatus.flyHeight;
		  write( GISstatus.procTab[GISnumber].sockfd, GISstatus.camera, CAMP_LEN );
		  cPrms[1] = height;
		}
	      else
		write( GISstatus.procTab[GISnumber].sockfd, GISstatus.camera, CAMP_LEN );
	    }
	}
      GISstatus.procTab[GISnumber].cont = TERRAIN;
      Tcl_SetVar( ( Tcl_Interp* )interp, GISnumber == 0 ? "leftcontents" : 
		  GISnumber == 1 ? "centercontents" : "rightcontents", "1", 0 );
#ifdef DEBUG
      fprintf( stderr, "recveived CTER for %d\n", GISnumber );
#endif
      break;

    case CMAP:
      readAll( civ_sock, buffer, CIVheader[1]-1 );
      GISnumber = buffer[0];
      if( GISstatus.procTab[GISnumber].stat != GIS_OFF )
	{
	  GISstatus.url = Tcl_GetVar( ( Tcl_Interp* )interp, "newurl", 0 );
	  if( *GISstatus.url )
	    {
	      compURL( GISstatus.url, MAP );
	      write( GISstatus.procTab[GISnumber].sockfd, buffer, buffer[1]+2 );

	      GISstatus.camera[0] = CAMP;
	      if( GISstatus.procTab[GISnumber].orient == FLY ) 
		{
		  height = cPrms[1];
		  cPrms[1] = GISstatus.flyHeight;
		  write( GISstatus.procTab[GISnumber].sockfd, GISstatus.camera, CAMP_LEN );
		  cPrms[1] = height;
		}
	      else
		write( GISstatus.procTab[GISnumber].sockfd, GISstatus.camera, CAMP_LEN );
	    }
	}
      GISstatus.procTab[GISnumber].cont = MAP;
      Tcl_SetVar( ( Tcl_Interp* )interp, GISnumber == 0 ? "leftcontents" : 
		  GISnumber == 1 ? "centercontents" : "rightcontents", "2", 0 );
#ifdef DEBUG
      fprintf( stderr, "recveived CMAP for %d\n", GISnumber );
#endif
      break;

    case CWEA:
      readAll( civ_sock, buffer, CIVheader[1]-1 );
      GISnumber = buffer[0];
      if( GISstatus.procTab[GISnumber].stat != GIS_OFF )
	{
	  GISstatus.url = Tcl_GetVar( ( Tcl_Interp* )interp, "newurl", 0 );
	  if( *GISstatus.url )
	    {
	      compURL( GISstatus.url, WEATHER );
	      write( GISstatus.procTab[GISnumber].sockfd, buffer, buffer[1]+2 );

	      GISstatus.camera[0] = CAMP;
	      if( GISstatus.procTab[GISnumber].orient == FLY ) 
		{
		  height = cPrms[1];
		  cPrms[1] = GISstatus.flyHeight;
		  write( GISstatus.procTab[GISnumber].sockfd, GISstatus.camera, CAMP_LEN );
		  cPrms[1] = height;
		}
	      else
		write( GISstatus.procTab[GISnumber].sockfd, GISstatus.camera, CAMP_LEN );
	    }
	}
      GISstatus.procTab[GISnumber].cont = WEATHER;
      Tcl_SetVar( ( Tcl_Interp* )interp, GISnumber == 0 ? "leftcontents" : 
		  GISnumber == 1 ? "centercontents" : "rightcontents", "3", 0 );
#ifdef DEBUG
      fprintf( stderr, "recveived CWEA for %d\n", GISnumber );
#endif
      break;

    case CWALK:
      read( civ_sock, &cbuf, 1 );
#ifdef DEBUG
      fprintf( stderr, "recveived CWALK for %d\n", cbuf );
#endif
      if( GISstatus.procTab[cbuf].orient != WALK )
	{
	  if( GISstatus.procTab[cbuf].stat != GIS_OFF )
	    {
#ifdef DEBUG
	      fprintf( stderr, "CWALK - CAMP:\t\t(%f, %f, %f)\n\t\t\t(%f, %f, %f, %f)\n",
		       cPrms[0], cPrms[1], cPrms[2], cPrms[3], cPrms[4], cPrms[5], cPrms[6] );
#endif
	      write( GISstatus.procTab[cbuf].sockfd, GISstatus.camera, CAMP_LEN );
	    }
	  GISstatus.procTab[cbuf].orient = WALK;
	  Tcl_SetVar( ( Tcl_Interp* )interp, cbuf == 0 ? "leftorient" :
		      cbuf == 1 ? "centerorient" : "rightorient", "4", 0 );
	}
      break;

    case CFLY:
      read( civ_sock, &cbuf, 1 );
#ifdef DEBUG
      fprintf( stderr, "recveived CFLY for %d\n", cbuf );
#endif

      if( GISstatus.procTab[cbuf].orient != FLY )
	{
	  if( GISstatus.procTab[cbuf].stat != GIS_OFF )
	    {
	      height = cPrms[1];
	      cPrms[1] = GISstatus.flyHeight;
#ifdef DEBUG
	      fprintf( stderr, "CFLY - CAMP:\t\t(%f, %f, %f)\n\t\t\t(%f, %f, %f, %f)\n",
		       cPrms[0], cPrms[1], cPrms[2], cPrms[3], cPrms[4], cPrms[5], cPrms[6] );
#endif
	      write( GISstatus.procTab[cbuf].sockfd, GISstatus.camera, CAMP_LEN );
	      cPrms[1] = height;
	    }
	  GISstatus.procTab[cbuf].orient = FLY;
	  Tcl_SetVar( ( Tcl_Interp* )interp, cbuf == 0 ? "leftorient" : 
		      cbuf == 1 ? "centerorient" : "rightorient", "5", 0 );
	}
      break;

    case CHGT:
      read( civ_sock, &cbuf, 1 );
      GISstatus.flyHeight = cbuf;
      for( GISnumber=1; GISnumber<PROCMAX; GISnumber++ )
	if( ( GISstatus.procTab[GISnumber].stat != GIS_OFF )&&
	    ( GISstatus.procTab[GISnumber].orient == FLY ) )
	  {
	    height = cPrms[1];
	    cPrms[1] = GISstatus.flyHeight;
	    GISstatus.camera[0] = CAMP;
	    write( GISstatus.procTab[GISnumber].sockfd, GISstatus.camera, CAMP_LEN );
	    cPrms[1] = height;
	  }
      //      char height_str[8];
      sprintf( buffer, "%d", cbuf );
      Tcl_SetVar( ( Tcl_Interp* )interp, "height", buffer, 0 );
      break;
      
    case SGIS:
      read( civ_sock, &cbuf, 1 );
      spawnGIS( cbuf, ( Tcl_Interp* )interp );
      Tcl_SetVar( ( Tcl_Interp* )interp, cbuf == 0 ? "leftstatus" : 
		  cbuf == 1 ? "centerstatus" : "rightstatus", "1", 0 );
      GISstatus.camera[0] = CAMP;
      if( GISstatus.procTab[cbuf].orient == FLY ) 
	{
	  height = cPrms[1];
	  cPrms[1] = GISstatus.flyHeight;
	  write( GISstatus.procTab[cbuf].sockfd, GISstatus.camera, CAMP_LEN );
	  cPrms[1] = height;
	}
      else
	  write( GISstatus.procTab[cbuf].sockfd, GISstatus.camera, CAMP_LEN );
#ifdef DEBUG
      fprintf( stderr, "---\nreceived CAMP:\t\t(%f, %f, %f)\n\t\t\t(%f, %f, %f, %f)\n",
	       cPrms[0], cPrms[1], cPrms[2], cPrms[3], cPrms[4], cPrms[5], cPrms[6] );
#endif
      break;

    case KGIS:
      read( civ_sock, &cbuf, 1 );
      killGIS( cbuf );
      Tcl_SetVar( ( Tcl_Interp* )interp, cbuf == 0 ? "leftstatus" : 
		  cbuf == 1 ? "centerstatus" : "rightstatus", "0", 0 );
      break;
      
    case SANS:
      //      fprintf( stderr, "SYNC ans received\n" );
      readAll( civ_sock, buffer, CIVheader[1]-1 );
      //      fprintf( stderr, "GISstatus read\n" );
      if( ! init )
	{
	  for( GISnumber=0; GISnumber<PROCMAX; GISnumber++ )
	    {
	      memcpy( &GISstatus.procTab[GISnumber].stat, buffer + GISnumber * 3, 3 );
	      Tcl_SetVar( ( Tcl_Interp* )interp,
			  GISnumber == 0 ? "leftstatus" : 
			  GISnumber == 1 ? "centerstatus" : "rightstatus", 
			  GISstatus.procTab[GISnumber].stat == GIS_OFF ? "0" : "1",
			  0 );
	      Tcl_SetVar( ( Tcl_Interp* )interp,
			  GISnumber == 0 ? "leftcontents" : 
			  GISnumber == 1 ? "centercontents" : "rightcontents", 
			  GISstatus.procTab[GISnumber].cont == TERRAIN ? "1" :
			  GISstatus.procTab[GISnumber].cont == MAP ? "2" : "3",
			  0 );
	      Tcl_SetVar( ( Tcl_Interp* )interp, 
			  GISnumber == 0 ? "leftorient" : 
			  GISnumber == 1 ? "centerorient" : "rightorient", 
			  GISstatus.procTab[GISnumber].orient == WALK ? "4" : "5",
			  0 );
	    }
	  //	  fprintf( stderr, "GISstatus set\n" );
	  for( GISnumber=0; GISnumber<PROCMAX; GISnumber++ )
	    if( GISstatus.procTab[GISnumber].stat != GIS_OFF )
	      spawnGIS( GISnumber, ( Tcl_Interp* )interp );
	  init = 1;
	}
      break;

    case NOBJ:
      //      fprintf( stderr, "NOBJ received\n" );
      readAll( civ_sock, &buffer[1], CIVheader[1]-1 );
      buffer[0] = NOBJ;
      
      break;
      
    case SREQ:
      break;

      //#ifdef DEBUG
    default:
      ;
      //      fprintf( stderr, "GIS protocol violation: %d\t", *GISstatus.camera );
      //      fprintf( stderr, "pruning of next %d bytes\n", read( civ_sock, buffer, CIVheader[1]-1 ) );
      //fprintf( stdout, "pruning of next %d bytes\n", read( civ_sock, buffer, URL_MAXLEN ) );
      //#endif
    }
}


int
startGIS_cmd( ClientData /*clientData*/,
	      Tcl_Interp *interp,
	      int /*argc*/, char *argv[])
{
  Tcl_GetInt( interp, argv[1], &GISnumber );
  if( app_mode == MST )
    {
      if( ( GISstatus.procTab[0].stat == GIS_OFF )&&( GISnumber != 0 ) )
	{
	  Tcl_SetVar( interp, GISnumber==1 ? "centerstatus" : "rightstatus", "0", 0 );
	  fprintf( stderr, "Start Master first\n" );
	  return TCL_OK;        
	}
  
      if( ( GISstatus.procTab[GISnumber].stat == GIS_OFF )&&( GISnumber < PROCMAX ) )
	{
	  spawnGIS( GISnumber, interp );
	  if( GISnumber != 0 )
	    {
	      GISstatus.camera[0] = CAMP;
	      if( GISstatus.procTab[GISnumber].orient == FLY ) 
		{
		  float* cPrms = ( float* )( GISstatus.camera+1 );
		  height = cPrms[1];
		  cPrms[1] = GISstatus.flyHeight;
		  write( GISstatus.procTab[GISnumber].sockfd, GISstatus.camera, CAMP_LEN );
		  cPrms[1] = height;
		}
	      else
		write( GISstatus.procTab[GISnumber].sockfd, GISstatus.camera, CAMP_LEN );
#ifdef DEBUG
	      fprintf( stderr, "---\nreceived CAMP:\t\t(%f, %f, %f)\n\t\t\t(%f, %f, %f, %f)\n",
		       cPrms[0], cPrms[1], cPrms[2], cPrms[3], cPrms[4], cPrms[5], cPrms[6] );
#endif
	    }

#ifdef DEBUG
	  fprintf( stderr, "sending SGIS\n" );
#endif
	  CIVheader[0] = DATA;
	  CIVheader[1] = 2;
	  buffer[0] = SGIS;
	  buffer[1] = ( char )GISnumber;
	  write( civ_sock, CIVheader, sizeof( CIVheader ) );
	  write( civ_sock, buffer, 2 );
	}
    }
  else
    if( GISstatus.procTab[GISnumber].stat == GIS_OFF )
      Tcl_SetVar( ( Tcl_Interp* )interp, GISnumber == 0 ? "leftstatus" : 
		  GISnumber == 1 ? "centerstatus" : "rightstatus", "0", 0 );
  return TCL_OK;  
}


int
stopGIS_cmd( ClientData /*clientData*/,
	     Tcl_Interp *interp,
	     int /*argc*/, char* argv[] )
{
  Tcl_GetInt( interp, argv[1], &GISnumber );

  if( app_mode == MST )
    {
      if( GISstatus.procTab[GISnumber].stat != GIS_OFF )
	{
	  killGIS( GISnumber );
#ifdef DEBUG
	  fprintf( stderr, "sending KGIS\n" );
#endif
	  CIVheader[0] = DATA;
	  CIVheader[1] = 2;
	  buffer[0] = KGIS;
	  buffer[1] = ( char )GISnumber;
	  write( civ_sock, CIVheader, sizeof( CIVheader ) );
	  write( civ_sock, buffer, 2 );
	}
    }
  else
    if( GISstatus.procTab[GISnumber].stat == GIS_ON )
      Tcl_SetVar( ( Tcl_Interp* )interp, GISnumber == 0 ? "leftstatus" : 
		  GISnumber == 1 ? "centerstatus" : "rightstatus", "1", 0 );
      
  return TCL_OK;  
}

int
sendNURL_cmd( ClientData /*clientData*/,
	      Tcl_Interp *interp,
	      int /*argc*/, char */*argv*/[] )
{
  if( app_mode == MST )
    {
      GISstatus.url = Tcl_GetVar( interp, "newurl", 0 );
      if( ! strncmp( GISstatus.url, "http://", 7/**GISstatus.url*/ ) )
	{
	  strncpy( &buffer[2], GISstatus.url, URL_MAXLEN-7 );
	  buffer[0] = NURL;
	  cbuf = strlen( &buffer[2] );
	  for( GISnumber=0; GISnumber<PROCMAX; GISnumber++ )
	    if( GISstatus.procTab[GISnumber].stat != GIS_OFF )
	      {
		strcpy( &buffer[2] + cbuf, GISstatus.procTab[GISnumber].cont == TERRAIN ? suffix[0]:
			GISstatus.procTab[GISnumber].cont == MAP ? suffix[1] : suffix[2] );

		buffer[1] = cbuf + 5; //strlen + suffix(4) +'\0'
#ifdef DEBUG
		fprintf( stderr, "sending NURL: %s, len %d\n", &buffer[2], buffer[1] );
#endif	    
		write( GISstatus.procTab[GISnumber].sockfd, buffer, buffer[1]+2 );
	      }

	  CIVheader[0] = DATA;
	  CIVheader[1] = strlen( GISstatus.url )+2; // NURL + strlen + '0'
	  cbuf = NURL;
#ifdef DEBUG
	  fprintf( stderr, "CIVheader: %d, %d\n", CIVheader[0], CIVheader[1] );
#endif
	  write( civ_sock, CIVheader, sizeof( CIVheader ) );
	  write( civ_sock, &cbuf, 1 );
	  write( civ_sock, GISstatus.url, CIVheader[1] - 1/*strlen( GISstatus.url )+1*/ );
	}
    }

  return TCL_OK;
}


int
displayTerrain_cmd( ClientData /*clientData*/, Tcl_Interp *interp,
		    int /*argc*/, char *argv[] )
{

  Tcl_GetInt( interp, argv[1], &GISnumber );
  if( ( app_mode == MST )&&
      ( GISstatus.procTab[GISnumber].cont != TERRAIN ) )
    {
      if( GISstatus.procTab[GISnumber].stat != GIS_OFF )
	{
	  GISstatus.url = Tcl_GetVar( interp, "newurl", 0 );
	  if( strlen( GISstatus.url ) != 0 )
	    {
	      compURL( GISstatus.url, TERRAIN );
	      write( GISstatus.procTab[GISnumber].sockfd, buffer, buffer[1]+2 );

	      float* cPrms = ( float* )( GISstatus.camera+1 );
	      height = cPrms[1];
	      GISstatus.camera[0] = CAMP;
	      if( GISstatus.procTab[GISnumber].orient == FLY ) 
		{
		  height = cPrms[1];
		  cPrms[1] = GISstatus.flyHeight;
		  write( GISstatus.procTab[GISnumber].sockfd, GISstatus.camera, CAMP_LEN );
		  cPrms[1] = height;
		}
	      else
		write( GISstatus.procTab[GISnumber].sockfd, GISstatus.camera, CAMP_LEN );
	    }
	}
	
      GISstatus.procTab[GISnumber].cont = TERRAIN;
      CIVheader[0] = DATA;
      CIVheader[1] = 2;
      buffer[0] = CTER;
      buffer[1] = ( char )GISnumber;
#ifdef DEBUG
      fprintf( stderr, "CIVheader: %d, %d\n", CIVheader[0], CIVheader[1] );
      fprintf( stderr, "Data: %d, %d\n", buffer[0], buffer[1] );
#endif
      write( civ_sock, CIVheader, sizeof( CIVheader ) );
      write( civ_sock, buffer, CIVheader[1] );
    }
  else 
    switch( GISstatus.procTab[GISnumber].cont )
      {
      case MAP:
	Tcl_SetVar( ( Tcl_Interp* )interp, GISnumber == 0 ? "leftcontents" : 
		    GISnumber == 1 ? "centercontents" : "rightcontents", "2", 0 );
	break;
      case WEATHER:
	Tcl_SetVar( ( Tcl_Interp* )interp, GISnumber == 0 ? "leftcontents" : 
		    GISnumber == 1 ? "centercontents" : "rightcontents", "3", 0 );
	break;
      }
  return TCL_OK;  
}


int
displayMap_cmd( ClientData /*clientData*/, Tcl_Interp *interp,
		int /*argc*/, char *argv[] )
{

  Tcl_GetInt( interp, argv[1], &GISnumber );
  if( ( app_mode == MST )&&
      ( GISstatus.procTab[GISnumber].cont != MAP ) )
    {
      if( GISstatus.procTab[GISnumber].stat != GIS_OFF )
	{
	  GISstatus.url = Tcl_GetVar( interp, "newurl", 0 );
	  if( strlen( GISstatus.url ) != 0 )
	    {
	      compURL( GISstatus.url, MAP );
	      write( GISstatus.procTab[GISnumber].sockfd, buffer, buffer[1]+2 );

	      float* cPrms = ( float* )( GISstatus.camera+1 );
	      height = cPrms[1];
	      GISstatus.camera[0] = CAMP;
	      if( GISstatus.procTab[GISnumber].orient == FLY ) 
		{
		  height = cPrms[1];
		  cPrms[1] = GISstatus.flyHeight;
		  write( GISstatus.procTab[GISnumber].sockfd, GISstatus.camera, CAMP_LEN );
		  cPrms[1] = height;
		}
	      else
		write( GISstatus.procTab[GISnumber].sockfd, GISstatus.camera, CAMP_LEN );
	    }
	}
      GISstatus.procTab[GISnumber].cont = MAP;
      CIVheader[0] = DATA;
      CIVheader[1] = 2;
      buffer[0] = CMAP;
      buffer[1] = ( char )GISnumber;
#ifdef DEBUG
      fprintf( stderr, "CIVheader: %d, %d\n", CIVheader[0], CIVheader[1] );
      fprintf( stderr, "Data: %d, %d\n", buffer[0], buffer[1] );
#endif
      write( civ_sock, CIVheader, sizeof( CIVheader ) );
      write( civ_sock, buffer, CIVheader[1] );
    }
  else 
    switch( GISstatus.procTab[GISnumber].cont )
      {
      case TERRAIN:
	Tcl_SetVar( ( Tcl_Interp* )interp, GISnumber == 0 ? "leftcontents" : 
		    GISnumber == 1 ? "centercontents" : "rightcontents", "1", 0 );
	break;
      case WEATHER:
	Tcl_SetVar( ( Tcl_Interp* )interp, GISnumber == 0 ? "leftcontents" : 
		    GISnumber == 1 ? "centercontents" : "rightcontents", "3", 0 );
	break;
      }
  return TCL_OK;  
}


int
displayWeather_cmd( ClientData /*clientData*/, Tcl_Interp *interp,
		    int /*argc*/, char *argv[] )
{

  Tcl_GetInt( interp, argv[1], &GISnumber );
  if( ( app_mode == MST )&&
      ( GISstatus.procTab[GISnumber].cont != WEATHER ) )
    {
      if( GISstatus.procTab[GISnumber].stat != GIS_OFF )
	{
	  GISstatus.url = Tcl_GetVar( interp, "newurl", 0 );
	  if( strlen( GISstatus.url ) != 0 )
	    {
	      compURL( GISstatus.url, WEATHER );
	      write( GISstatus.procTab[GISnumber].sockfd, buffer, buffer[1]+2 );

	      float* cPrms = ( float* )( GISstatus.camera+1 );
	      height = cPrms[1];
	      GISstatus.camera[0] = CAMP;
	      if( GISstatus.procTab[GISnumber].orient == FLY ) 
		{
		  height = cPrms[1];
		  cPrms[1] = GISstatus.flyHeight;
		  write( GISstatus.procTab[GISnumber].sockfd, GISstatus.camera, CAMP_LEN );
		  cPrms[1] = height;
		}
	      else
		write( GISstatus.procTab[GISnumber].sockfd, GISstatus.camera, CAMP_LEN );
	    }
	}
      GISstatus.procTab[GISnumber].cont = WEATHER;
      CIVheader[0] = DATA;
      CIVheader[1] = 2;
      buffer[0] = CWEA;
      buffer[1] = ( char )GISnumber;
#ifdef DEBUG
      fprintf( stderr, "CIVheader: %d, %d\n", CIVheader[0], CIVheader[1] );
      fprintf( stderr, "Data: %d, %d\n", buffer[0], buffer[1] );
#endif
      write( civ_sock, CIVheader, sizeof( CIVheader ) );
      write( civ_sock, buffer, CIVheader[1] );
    }
  else 
    switch( GISstatus.procTab[GISnumber].cont )
      {
      case TERRAIN:
	Tcl_SetVar( ( Tcl_Interp* )interp, GISnumber == 0 ? "leftcontents" : 
		    GISnumber == 1 ? "centercontents" : "rightcontents", "1", 0 );
	break;
      case MAP:
	Tcl_SetVar( ( Tcl_Interp* )interp, GISnumber == 0 ? "leftcontents" : 
		    GISnumber == 1 ? "centercontents" : "rightcontents", "2", 0 );
	break;
      }
  return TCL_OK;  
}


int
showAppInfo_cmd(ClientData /*clientData*/,
		Tcl_Interp */*interp*/,
		int /*argc*/, char */*argv*/[])
{
  fprintf ( stderr, "\n" );
  for( GISnumber=0; GISnumber<PROCMAX; GISnumber++ )
    fprintf( stderr, "GIS: %d, proc: %d, sockfd: %d, contents: %s, orient: %s\n",
	     GISnumber, GISstatus.procTab[GISnumber].pid, GISstatus.procTab[GISnumber].sockfd, 
	     GISstatus.procTab[GISnumber].cont == TERRAIN ? "terrain" : 
	     GISstatus.procTab[GISnumber].cont == MAP ? "map" : "weather",
	     GISstatus.procTab[GISnumber].orient == WALK ? "walk" : "fly" );

  return TCL_OK;  
}


int
quitGISapp_cmd(ClientData /*clientData*/,
	       Tcl_Interp */*interp*/,
	       int /*argc*/, char */*argv*/[])
{
  Tk_DeleteFileHandler( GISstatus.procTab[0].sockfd );
  if( app_mode == SLV )
    Tk_DeleteFileHandler( civ_sock );
  close( civ_sock );
  
  for( GISnumber=0; GISnumber<PROCMAX; GISnumber++ )
    if( GISstatus.procTab[GISnumber].stat != GIS_OFF )
      {
	kill( GISstatus.procTab[GISnumber].pid, SIGTERM );
	close( GISstatus.procTab[GISnumber].sockfd );
      }
  
  return TCL_OK;  
}

int
switchToWalk_cmd( ClientData /*clientData*/, Tcl_Interp *interp,
		  int /*argc*/, char *argv[] )
{
  Tcl_GetInt( interp, argv[1], &GISnumber );
  if( app_mode == MST )
    {
      if( GISstatus.procTab[GISnumber].orient != WALK )
	{
	  if( GISstatus.procTab[GISnumber].stat != GIS_OFF )
	    {
	      GISstatus.camera[0] = CWALK;
	      write( GISstatus.procTab[GISnumber].sockfd, GISstatus.camera, CAMP_LEN );
	    }
	  GISstatus.procTab[GISnumber].orient = WALK;
	  CIVheader[0] = DATA;
	  CIVheader[1] = 2;
	  buffer[0] = CWALK;
	  buffer[1] = ( char )GISnumber;
#ifdef DEBUG
	  fprintf( stderr, "CIVheader: %d, %d\n", CIVheader[0], CIVheader[1] );
	  fprintf( stderr, "Data: %d, %d\n", buffer[0], buffer[1] );
#endif
	  write( civ_sock, CIVheader, sizeof( CIVheader ) );
	  write( civ_sock, buffer, 2 );
	}
    }
  else 
    if( GISstatus.procTab[GISnumber].orient == FLY )
      Tcl_SetVar( ( Tcl_Interp* )interp, GISnumber == 0 ? "leftorient" : 
		  GISnumber == 1 ? "centerorient" : "rightorient", "5", 0 );
  return TCL_OK;  
}



int
switchToFly_cmd( ClientData /*clientData*/, Tcl_Interp *interp,
		 int /*argc*/, char *argv[] )
{
  Tcl_GetInt( interp, argv[1], &GISnumber );
  if( app_mode == MST )
    {
      if( GISstatus.procTab[GISnumber].orient != FLY )
	{
	  if( GISstatus.procTab[GISnumber].stat != GIS_OFF )
	    {
	      float* cPrms = ( float* )( GISstatus.camera+1 );
	      height = cPrms[1];
	      cPrms[1] = GISstatus.flyHeight;
	      GISstatus.camera[0] = CFLY;
	      write( GISstatus.procTab[GISnumber].sockfd, GISstatus.camera, CAMP_LEN );
	      cPrms[1] = height;
	    }
	  GISstatus.procTab[GISnumber].orient = FLY;
	  CIVheader[0] = DATA;
	  CIVheader[1] = 2;
	  buffer[0] = CFLY;
	  buffer[1] = ( char )GISnumber;
#ifdef DEBUG
	  fprintf( stderr, "CIVheader: %d, %d\n", CIVheader[0], CIVheader[1] );
	  fprintf( stderr, "Data: %d, %d\n", buffer[0], buffer[1] );
#endif
	  write( civ_sock, CIVheader, sizeof( CIVheader ) );
	  write( civ_sock, buffer, 2 );
	}
    }
  else 
    if( GISstatus.procTab[GISnumber].orient == WALK )
      Tcl_SetVar( ( Tcl_Interp* )interp, GISnumber == 0 ? "leftorient" : 
		  GISnumber == 1 ? "centerorient" : "rightorient", "4", 0 );
  return TCL_OK;  
}

int
setHeight_cmd( ClientData /*clientData*/, Tcl_Interp *interp,
		 int /*argc*/, char *argv[] )
{
  if( app_mode == MST )
    {
      Tcl_GetInt( interp, argv[1], &GISnumber );
      GISstatus.flyHeight = GISnumber;
      
      CIVheader[0] = DATA;
      CIVheader[1] = 2;
      buffer[0] = CHGT;
      buffer[1] = ( char )GISnumber;
#ifdef DEBUG
      fprintf( stderr, "CIVheader: %d, %d\n", CIVheader[0], CIVheader[1] );
      fprintf( stderr, "Data: %d, %d\n", buffer[0], buffer[1] );
#endif
      write( civ_sock, CIVheader, sizeof( CIVheader ) );
      write( civ_sock, buffer, 2 );
      
      for( GISnumber=1; GISnumber<PROCMAX; GISnumber++ )
	if( ( GISstatus.procTab[GISnumber].stat != GIS_OFF )&&
	    ( GISstatus.procTab[GISnumber].orient == FLY ) )
	  {
	    float* cPrms = ( float* )( GISstatus.camera+1 );
	    height = cPrms[1];
	    cPrms[1] = GISstatus.flyHeight;
	    GISstatus.camera[0] = CAMP;
	    write( GISstatus.procTab[GISnumber].sockfd, GISstatus.camera, CAMP_LEN );
	    cPrms[1] = height;
	  }
    } 
  else
    {
      sprintf( buffer, "%d", ( char )GISstatus.flyHeight );
      Tcl_SetVar( ( Tcl_Interp* )interp, "height", buffer, 0 );
    }
  
  return TCL_OK;  
}

#define BEAST_ADDR "128.230.162.22"

int
Tcl_AppInit( Tcl_Interp *interp )
{
  if ( Tcl_Init(interp) == TCL_ERROR )
    return TCL_ERROR;

  if ( Tk_Init(interp) == TCL_ERROR )
    return TCL_ERROR;

  Tcl_CreateCommand(interp, "startGIS", startGIS_cmd,
		    (ClientData)Tk_MainWindow(interp), (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateCommand(interp, "stopGIS", stopGIS_cmd,
		    (ClientData)Tk_MainWindow(interp), (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateCommand(interp, "quitGISapp", quitGISapp_cmd,
		    (ClientData)Tk_MainWindow(interp), (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateCommand(interp, "showAppInfo", showAppInfo_cmd,
		    (ClientData)Tk_MainWindow(interp), (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateCommand(interp, "sendNURL", sendNURL_cmd,
		    (ClientData)Tk_MainWindow(interp), (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateCommand(interp, "displayTerrain", displayTerrain_cmd,
		    (ClientData)Tk_MainWindow(interp), (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateCommand(interp, "displayMap", displayMap_cmd,
		    (ClientData)Tk_MainWindow(interp), (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateCommand(interp, "displayWeather", displayWeather_cmd,
		    (ClientData)Tk_MainWindow(interp), (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateCommand(interp, "switchToWalk", switchToWalk_cmd,
		    (ClientData)Tk_MainWindow(interp), (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateCommand(interp, "switchToFly", switchToFly_cmd,
		    (ClientData)Tk_MainWindow(interp), (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateCommand(interp, "setHeight", setHeight_cmd,
		    (ClientData)Tk_MainWindow(interp), (Tcl_CmdDeleteProc *)NULL);

  for( GISnumber=0; GISnumber<PROCMAX; GISnumber++ )
    {
      GISstatus.procTab[GISnumber].stat = GIS_OFF;
      GISstatus.procTab[GISnumber].cont = TERRAIN;
      GISstatus.procTab[GISnumber].orient = WALK;
    }

  GISstatus.flyHeight = 60;
  GISstatus.url = NULL;
  init = 0; 
  
#ifdef CIV_MODE  
  civ_sock = registerApp( AID, port );
#else
  struct sockaddr_in name;
  if( ( civ_sock = socket( PF_INET, SOCK_STREAM, 0 ) ) < 0 )
    {
      perror( "socket" );
      exit( 1 );
    }

  name.sin_family = AF_INET;
  name.sin_port = htons( port );

  struct hostent* hostinfo;
  hostinfo = gethostbyname( trans_addr );
  if( ! hostinfo )
    {
      fprintf( stderr, "%s: %s\n", trans_addr, hstrerror( h_errno ) );
      exit( 1 );
    }
  name.sin_addr = *( struct in_addr* )hostinfo->h_addr;
  if( connect( civ_sock, ( struct sockaddr* )&name, sizeof( name ) ) < 0 )
    {
      perror( "connect" );
      exit( 1 );
    }
#endif

#ifdef DEBUG
  fprintf( stderr, "Client connected\n" );
  fprintf( stderr, "civ_sock: %d\n", civ_sock );
#endif

  char isMaster[2] = { 1 };
  CIVheader[0] = CONTROL;
  CIVheader[1] = 1;
  write( civ_sock, CIVheader, sizeof( CIVheader ) );
  write( civ_sock, isMaster, 1 );
  
  readAll( civ_sock, CIVheader, sizeof( CIVheader ) );
  readAll( civ_sock, isMaster, 2 );

#ifdef DEBUG
  fprintf( stderr, "CIVheader: %d, %d\n", CIVheader[0], CIVheader[1] );
  fprintf( stderr, "App mode: %s\n", isMaster[1] == 10 ? "master" : "slave" );
#endif

  if( isMaster[1] == 10 )
    {
      app_mode = MST;
      Tk_CreateFileHandler( civ_sock, TK_READABLE, readFromContrApp, interp );
    }
  else
    { 
      app_mode = SLV; 
      Tk_CreateFileHandler( civ_sock, TK_READABLE, readFromMasterApp, interp );
      CIVheader[0] = DATA;
      CIVheader[1] = 1;
      cbuf = SREQ;
      write( civ_sock, CIVheader, sizeof( CIVheader ) );
      write( civ_sock, &cbuf, 1 );
    } 

#ifdef CIV_MODE  
  Tcl_EvalFile(interp, "/project/K136/CIV/CIVGIS/Browser/pooh.tcl" );
#else
  Tcl_EvalFile(interp, "pooh.tcl" );
#endif
  return TCL_OK;
}
